@@ -181,8 +181,8 @@ class FieldsInvalidFieldError(Error):
         }
 
 
-class FilterInvalidFieldError(Error):
-    """Errors related to an invalid filter field."""
+class FilterInvalidPathError(Error):
+    """Errors related to an invalid filter path."""
 
     # pylint: disable=too-few-public-methods
 
@@ -198,7 +198,29 @@ class FilterInvalidFieldError(Error):
                 'parameter': 'filter'
             },
             'title': 'Invalid query parameter.',
-            'detail': f"The resource does not have a '{self.path}' field."
+            'detail': f"The resource does not have a '{self.path}' path."
+        }
+
+
+class FilterInvalidValueError(Error):
+    """Errors related to an invalid filter value."""
+
+    # pylint: disable=too-few-public-methods
+
+    status = 400
+
+    def __init__(self, path, value):
+        self.path = path
+        self.value = value
+
+    def to_error(self):
+        return {
+            'status': '400',
+            'source': {
+                'parameter': 'filter'
+            },
+            'title': 'Invalid query parameter.',
+            'detail': f"Invalid value {self.value} for '{self.path}' field."
         }
 
 
