@@ -33,15 +33,16 @@ from asyncpg.exceptions import (
     RestrictViolationError,
     UniqueViolationError,
 )
+from multidict import MultiDict
 
-from ajsonapi.conversions import id_number_to_name
+from ajsonapi.conversions import id_name_to_number, id_number_to_name
 from ajsonapi.document import (
     ResourceObjectDocument,
     parse_request_document_data,
 )
 from ajsonapi.environments import GetCollectionIncludedEnvironment
 from ajsonapi.errors import (
-    FilterInvalidFieldError,
+    FilterInvalidPathError,
     IncludeInvalidPathError,
     QueryParameterUnsupportedError,
     ResourceNotFoundError,
@@ -63,7 +64,55 @@ from ajsonapi.query import (
 from ajsonapi.responses import document_response, no_content_response
 from ajsonapi.uri.resource import Resource
 
-RE_FILTER = re_compile(r'filter\[([^\]]*)\]')
+RE_FILTER = re_compile(r'filter%5B([^%]*)%5D')
+
+
+class FilterNode:
+    """A FilterNode is a node in the 'field' tree from a resource.
+    It contains values for fields as well as paths to deeper fields.
+    """
+
+    # pylint: disable=too-few-public-methods
+
+    def __init__(self, attributes=None, relationships=None):
+        if attributes is None:
+            attributes = {}
+        if relationships is None:
+            relationships = {}
+        self.attributes = attributes
+        self.relationships = relationships
+
+    def __str__(self):
+        rels_str = {
+            key: str(value) for key, value in self.relationships.items()
+        }
+        return f'FilterNode({self.attributes}, {rels_str})'
+
+    def conditions(self, collection):
+        """Returns the SQL condition for this filter node."""
+
+        conds = []
+        for rel, sub_node in self.relationships.items():
+            sub_col = rel.rtable.collection
+            sub_conds = sub_node.conditions(sub_col)
+            where_clause = f'WHERE {" AND ".join(sub_conds)}'
+            conds.append(
+                rel.filter_condition(
+                    f'SELECT id FROM data.{rel.rtable.name} {where_clause}'))
+        for attr_name, values in self.attributes.items():
+            if attr_name == 'id':
+                id_values = {f"'{value}'" for value in values}
+                conds.append(f'id IN ({",".join(id_values)})')
+            else:
+                attr = getattr(collection.table.___, attr_name)
+                conds.append(f'{attr_name} IN ({attr.filter_values(values)})')
+        return conds
+
+    def remote(self, relationship):
+        """Returns the remote FilterNode."""
+        if relationship in self.relationships:
+            return self.relationships[relationship]
+        return None
 
 
 class Collection(Resource):
@@ -111,7 +160,12 @@ class Collection(Resource):
         query = {}
         errors = []
 
-        request_query = request.query
+        parts = request.raw_path.split('?', 1)
+        if len(parts) == 1:
+            request_query = MultiDict()
+        else:
+            qps = parts[1]
+            request_query = MultiDict(qp.split('=', 1) for qp in qps.split('&'))
         self.parse_include(request_query, query, errors, allow=allow_include)
         parse_fields(request_query, query, errors, allow=allow_fields)
         self.parse_filter(request_query, query, errors, allow=allow_filter)
@@ -139,11 +193,19 @@ class Collection(Resource):
     def parse_filter(self, request_query, query, errors, allow):
         """Parses the request's filter query parameter."""
 
-        if query_has_parameter(request_query, 'filter'):
-            if allow:
-                query['filters'] = self.filters_list(request_query, errors)
-            else:
-                errors.append(QueryParameterUnsupportedError('filter'))
+        filter_path_values = []
+        for key, value in request_query.items():
+            match = RE_FILTER.match(key)
+            if match:
+                if not allow:
+                    errors.append(QueryParameterUnsupportedError('filter'))
+                    return
+                path = match.group(1)
+                filter_values = set(value.split(','))
+                filter_path_values.append((path, filter_values))
+        if filter_path_values != []:
+            query['filters'] = self.filters_tree(FilterNode(),
+                                                 filter_path_values, errors, [])
         else:
             query['filters'] = None
 
@@ -173,23 +235,71 @@ class Collection(Resource):
                     sub_include_dict, sub_includes, errors, path)
             path.pop()
 
-    def filters_list(self, request_query, errors):
+    def filters_tree(self, node, filter_path_values, errors, path):
         """Helper function for parsing request query filter parameters."""
 
-        return_list = []
-        for key, value in request_query.items():
-            match = RE_FILTER.match(key)
-            if match:
-                field_name = match.group(1)
-                try:
+        # pylint: disable=too-many-locals,too-many-branches
+
+        for (filter_path, filter_values) in filter_path_values:
+            field_name, *sub_filter_path = filter_path.split('.', 1)
+            path.append(field_name)
+            try:
+                if field_name != 'id':
                     field = getattr(self.table.___, field_name)
-                except AttributeError:
-                    errors.append(FilterInvalidFieldError(field_name))
-                    continue
                 else:
-                    condition = field.filter_condition(value)
-                return_list.append(condition)
-        return return_list
+                    field = None
+            except AttributeError:
+                errors.append(FilterInvalidPathError('.'.join(path)))
+            else:
+                from ajsonapi.attribute import \
+                    Attribute  # pylint: disable=import-outside-toplevel
+                from ajsonapi.relationships import \
+                    Relationship  # pylint: disable=import-outside-toplevel
+                if sub_filter_path == []:
+                    if isinstance(field, Relationship):
+                        # For backward compatibility, should become obsolete
+                        sub_node = node.relationships.get(field, FilterNode())
+                        filter_ids = set(
+                            id_name_to_number(filter_id)
+                            for filter_id in filter_values)
+                        ids = sub_node.attributes.get('id', None)
+                        if ids is None:
+                            sub_node.attributes['id'] = filter_ids
+                        else:
+                            sub_node.attributes['id'] = ids.intersection(
+                                filter_ids)
+                        node.relationships[field] = sub_node
+                    elif isinstance(field, Attribute):
+                        filter_attrs = set(
+                            field.type_.from_qp(v) for v in filter_values)
+                        values = node.attributes.get(field_name, None)
+                        if values is None:
+                            node.attributes[field_name] = filter_attrs
+                        else:
+                            node.attributes[field_name] = values.intersection(
+                                filter_attrs)
+                    elif field_name == 'id':
+                        filter_ids = set(f"'{id_name_to_number(filter_id)}'"
+                                         for filter_id in filter_values)
+                        ids = node.attributes.get('id', None)
+                        if ids is None:
+                            node.attributes['id'] = filter_ids
+                        else:
+                            node.attributes['id'] = ids.intersection(filter_ids)
+                    else:
+                        errors.append(FilterInvalidPathError('.'.join(path)))
+                else:
+                    if isinstance(field, Relationship):
+                        sub_col = field.rtable.collection
+                        sub_node = node.relationships.setdefault(
+                            field, FilterNode())
+                        sub_col.filters_tree(
+                            sub_node, [(*sub_filter_path, filter_values)],
+                            errors, path)
+                    else:
+                        errors.append(FilterInvalidPathError('.'.join(path)))
+            path.pop()
+        return node
 
     def sort_list(self, request_query, errors):
         """Helper function for parsing request query sort parameter."""
@@ -420,7 +530,7 @@ class Collection(Resource):
 
         filters = env.query['filters']
         if filters:
-            where_clause = f' WHERE {" AND ".join(filters)}'
+            where_clause = f' WHERE {" AND ".join(filters.conditions(self))}'
         else:
             where_clause = ''
 
